#include <assert.h>
#include <functional>
#include <memory>
#include <stdexcept>
#include <utility>

// Forward Declarations
template<class T, size_t Density> class ImmutableVector;

// Immutable Vector
template<class T, size_t Density> class ImmutableVector
{
public:
	typedef T ElementType;
	typedef ImmutableVector<T, Density> ThisType;
	enum { NumChildrenPerLevel = Density };

private:
	// Private Helper Objects
	struct Leaf
	{
		ElementType element_;
		bool fullyConstructed_;
		Leaf() : element_(ElementType()), fullyConstructed_(true) {}
		Leaf(ElementType element, bool locked = true) : element_(element), fullyConstructed_(locked) {}
		Leaf(const Leaf& rhs) : element_(rhs.element_), fullyConstructed_(rhs.fullyConstructed_) {}
		~Leaf() {}
		void SetFullyConstructed() { fullyConstructed_ = true; }
		bool IsFullyConstructed() const { return fullyConstructed_; }
		void SetValue(const ElementType& newValue) { assert(!fullyConstructed_); element_ = newValue; }
		const ElementType& GetValue() { return element_; }
	};
	struct Node
	{
		bool fullyConstructed_;
		Node() : fullyConstructed_(false) {}
		Node(const Node& rhs) : fullyConstructed_(rhs.fullyConstructed_) {}
		~Node() {}
		bool IsFullyConstructed() const { return fullyConstructed_; }
		virtual void SetFullyConstructed() = 0;
		virtual ElementType& GetElement(const size_t index, const size_t numElementsUnderThisNode) = 0;
		virtual const ElementType& GetValue(const size_t index, const size_t numElementsUnderThisNode) const = 0;
		virtual void SetValue(const size_t index, const size_t numElementsUnderThisNode, const ElementType &newValue) = 0;
	};
	struct LeafNode : public Node
	{
		std::shared_ptr<Leaf> children_[NumChildrenPerLevel];
		LeafNode() : Node() {}
		LeafNode(const LeafNode& rhs) : Node(rhs)
		{
			for (size_t child = 0; child < NumChildrenPerLevel; ++child) {
				children_[child] = rhs.children_[child];
			}			
		}
		virtual ~LeafNode() {}
		virtual void SetFullyConstructed()
		{
			for (size_t child = 0; child < NumChildrenPerLevel; ++child) {
				if ((nullptr != children_[child]) && (! children_[child]->IsFullyConstructed())) {
					children_[child]->SetFullyConstructed();
				}
			}
			fullyConstructed_ = true;
		}
		virtual ElementType& GetElement(const size_t index, const size_t /*numElementsUnderThisNode*/)
		{
			if (nullptr == children_[index]) {
				children_[index].reset(new Leaf(ElementType(), false));
			} else if (children_[index]->IsFullyConstructed()) {
				children_[index].reset(new Leaf(children_[index]->element_, false));
			}
			return children_[index]->element_;
		}
		virtual void SetValue(const size_t index, const size_t /*numElementsUnderThisNode*/, const ElementType &newValue)
		{
			if (nullptr == children_[index]) {
				children_[index].reset(new Leaf(newValue, false));
			} else if (children_[index]->IsFullyConstructed()) {
				children_[index].reset(new Leaf(newValue, false));
			} else {
				children_[index]->SetValue(newValue);
			}
		}
		virtual const ElementType& GetValue(const size_t index, const size_t /*numElementsUnderThisNode*/) const
		{
			if (nullptr == children_[index]) {
				throw std::invalid_argument("Invalid Child Node");
			} else {
				return children_[index]->GetValue();
			}
		}
	};
	struct IntermediateNode : public Node
	{
		std::shared_ptr<Node> children_[NumChildrenPerLevel];
		IntermediateNode() : Node() {}
		IntermediateNode(const IntermediateNode& rhs) : Node(rhs)
		{
			for (size_t child = 0; child < NumChildrenPerLevel; ++child) {
				children_[child] = rhs.children_[child];
			}			
		}
		virtual ~IntermediateNode() {}
		virtual void SetFullyConstructed()
		{
			for (size_t child = 0; child < NumChildrenPerLevel; ++child) {
				if ((nullptr != children_[child]) && (! children_[child]->IsFullyConstructed())) {
					children_[child]->SetFullyConstructed();
				}
			}
			fullyConstructed_ = true;
		}
		virtual ElementType& GetElement(const size_t index, const size_t numElementsUnderThisNode)
		{
			const size_t elementsPerChild = numElementsUnderThisNode / NumChildrenPerLevel;
			const size_t childIndex = index / elementsPerChild;
			const size_t childOffset = index % elementsPerChild;

			if (nullptr == children_[childIndex]) {
				if (NumChildrenPerLevel == elementsPerChild) {
					children_[childIndex].reset(new LeafNode());
				} else {
					children_[childIndex].reset(new IntermediateNode());
				}
				children_[childIndex]->fullyConstructed_ = false;
			} else if (children_[index]->IsFullyConstructed()) {
				if (NumChildrenPerLevel == elementsPerChild) {
					children_[childIndex].reset(new LeafNode(*(std::static_pointer_cast<LeafNode>(children_[index]))));
				} else {
					children_[childIndex].reset(new IntermediateNode(*(std::static_pointer_cast<IntermediateNode>(children_[index]))));
				}
				children_[childIndex]->fullyConstructed_ = false;
			}

			return children_[childIndex]->GetElement(childOffset, elementsPerChild);
		}
		virtual void SetValue(const size_t index, const size_t numElementsUnderThisNode, const ElementType &newValue)
		{
			const size_t elementsPerChild = numElementsUnderThisNode / NumChildrenPerLevel;
			const size_t childIndex = index / elementsPerChild;
			const size_t childOffset = index % elementsPerChild;

			if (nullptr == children_[childIndex]) {
				if (NumChildrenPerLevel == elementsPerChild) {
					children_[childIndex].reset(new LeafNode());
				} else {
					children_[childIndex].reset(new IntermediateNode());
				}
				children_[childIndex]->fullyConstructed_ = false;
			} else if (children_[childIndex]->IsFullyConstructed()) {
				if (NumChildrenPerLevel == elementsPerChild) {
					children_[childIndex].reset(new LeafNode(*(std::static_pointer_cast<LeafNode>(children_[childIndex]))));
				} else {
					children_[childIndex].reset(new IntermediateNode(*(std::static_pointer_cast<IntermediateNode>(children_[childIndex]))));
				}
				children_[childIndex]->fullyConstructed_ = false;
			}

			children_[childIndex]->SetValue(childOffset, elementsPerChild, newValue);
		}

		virtual const ElementType& GetValue(const size_t index, const size_t numElementsUnderThisNode) const
		{
			const size_t elementsPerChild = numElementsUnderThisNode / NumChildrenPerLevel;
			const size_t childIndex = index / elementsPerChild;
			const size_t childOffset = index % elementsPerChild;

			return children_[childIndex]->GetValue(childOffset, numElementsUnderThisNode);;
		}
	};

	// Data Members
	size_t capacity_;
	size_t size_;
	size_t numLevels_;
	std::shared_ptr<Node> rootNode_;

	ElementType& GetElement(const size_t index)
	{
		if (index >= size_)  throw std::invalid_argument("Invalid index");
		if (nullptr == rootNode_) {
			if (1 == numLevels_) {
				rootNode_.reset(new LeafNode());
			} else {
				rootNode_.reset(new IntermediateNode());
			}
			rootNode_->fullyConstructed_ = false;
		} else if (rootNode_->IsFullyConstructed()) {
			if (1 == numLevels_) {
				rootNode_.reset(new LeafNode(*(std::static_pointer_cast<LeafNode>(rootNode_))));
			} else {
				rootNode_.reset(new IntermediateNode(*(std::static_pointer_cast<IntermediateNode>(rootNode_))));
			}
			rootNode_->fullyConstructed_ = false;
		}
		return rootNode_->GetElement(index, capacity_);
	}

    void Reserve(const size_t capacity)
    {
        if (capacity <= capacity_) return;
        if (0 == capacity_) {
            if (capacity > 0) {
                numLevels_ = 1;
                capacity_ = NumChildrenPerLevel;
            }
            while (capacity_ < capacity) {
                ++numLevels_;
                capacity_ *= NumChildrenPerLevel;
            }
        } else {
            while (capacity_ < capacity) {
                ++numLevels_;
                capacity_ *= NumChildrenPerLevel;
                auto newRoot = std::make_shared<IntermediateNode>();
                newRoot->children_[0] = rootNode_;
                rootNode_ = newRoot;
            }
        }
    }

	void InternalSetValue(const size_t index, const ElementType& newValue)
	{
		if (index >= size_)  throw std::invalid_argument("Invalid index");
		if (nullptr == rootNode_) {
			if (1 == numLevels_) {
				rootNode_.reset(new LeafNode());
			} else {
				rootNode_.reset(new IntermediateNode());
			}
			rootNode_->fullyConstructed_ = false;
		} else if (rootNode_->IsFullyConstructed()) {
			if (1 == numLevels_) {
				rootNode_.reset(new LeafNode(*(std::static_pointer_cast<LeafNode>(rootNode_))));
			} else {
				rootNode_.reset(new IntermediateNode(*(std::static_pointer_cast<IntermediateNode>(rootNode_))));
			}
			rootNode_->fullyConstructed_ = false;
		}
		rootNode_->SetValue(index, capacity_, newValue);
	}

public:
	class MutableVector
	{
	private:
		ImmutableVector immutableVector_;

		MutableVector(const MutableVector&); // Do Not Allow Copying
		MutableVector& operator=(const MutableVector& rhs);

		// Do Not allow direct creation of MutableVector
        // (only ImmutableVector and other friends can create a new MutableVector)
		MutableVector(const ImmutableVector& immutableVector) : immutableVector_(immutableVector) {}
	public:
		friend class ImmutableVector<T, Density>;
		template<class T, size_t Density, class F> friend ImmutableVector<T, Density> GenerateFrom(const ImmutableVector<T, Density>& originalVector, F changeFunction);

		size_t Size() const { return immutableVector_.Size(); }
		void Reserve(const size_t capacity) { immutableVector_.Reserve(capacity); }
		ElementType& operator[](size_t index) {	return immutableVector_.GetElement(index);	}
		void PushBack(const ElementType& newValue) {
			const size_t size = immutableVector_.size_;
			const size_t newIndex = size;
			const size_t newSize = size + 1;
			immutableVector_.Reserve(newSize);
			++immutableVector_.size_;
			immutableVector_.InternalSetValue(newIndex, newValue);
		}
		ElementType PopBack() {
			const size_t size = immutableVector_.size_;
			if (0 == size) throw std::invalid_argument("Invalid Index");
			const size_t oldIndex = size - 1;
			ElementType oldValue = immutableVector_[oldIndex];
#ifdef _DEBUG
			immutableVector_.InternalSetValue(oldIndex, ElementType());
#endif
			--immutableVector_.size_;
			return oldValue;
		}
	};

	typedef MutableVector MutableCounterpart;

	ImmutableVector(const size_t capacity = 0) 
		: capacity_(0), size_(0), numLevels_(0)
	{
		if (capacity > 0)
		{
			numLevels_ = 1;
			capacity_ = NumChildrenPerLevel;
		}
		while (capacity_ < capacity) {
			++numLevels_;
			capacity_ *= NumChildrenPerLevel;
		}
	}

	ImmutableVector(const ImmutableVector&& rhs) 
		: capacity_(rhs.capacity_), size_(rhs.size_), numLevels_(rhs.numLevels_), rootNode_(std::move(rhs.rootNode_))
	{
	}

	ImmutableVector(const ImmutableVector& rhs) 
		: capacity_(rhs.capacity_), size_(rhs.size_), numLevels_(rhs.numLevels_), rootNode_(rhs.rootNode_)
	{
	}

	ImmutableVector& operator=(const ImmutableVector& rhs) 
	{
		capacity_ = rhs.capacity_;
		size_ = rhs.size_;
		numLevels_ = rhs.numLevels_;
		rootNode_ = rhs.rootNode_;
		return *this;
	}

	ImmutableVector(MutableVector&& rhs) 
		: capacity_(0), size_(0), numLevels_(0), rootNode_(nullptr)
	{
		if (this != &rhs.immutableVector_) {
			std::swap(capacity_, rhs.immutableVector_.capacity_);
			std::swap(size_, rhs.immutableVector_.size_);
			std::swap(numLevels_, rhs.immutableVector_.numLevels_);
			std::swap(rootNode_, rhs.immutableVector_.rootNode_);
		}
	}

	size_t Size() const
	{
		return size_;
	}

	const ElementType& operator[](size_t index) const
	{
		if (nullptr == rootNode_) {
            throw std::exception("Empty Vector");
		} else {
			return rootNode_->GetValue(index, capacity_);
		}
	}

    const ElementType& Front() const
    {
        if (nullptr == rootNode_) {
            throw std::exception("Empty Vector");
        } else {
            return rootNode_->GetValue(0, capacity_);
        }
    }

	const ElementType& Back() const
	{
		if (nullptr == rootNode_) {
			throw std::exception("Empty Vector");
		} else {
			return rootNode_->GetValue(size_ - 1, capacity_);
		}
	}

public:
	// Friends
	template<class T, size_t Density> friend ImmutableVector<T, Density> AppendValue(const ImmutableVector<T, Density>& v, const T& newValue);
	template<class T, size_t Density> friend ImmutableVector<T, Density> TruncateValue(const ImmutableVector<T, Density>& originalVector);
	template<class T, size_t Density> friend ImmutableVector<T, Density> SubstituteValueAtIndex(const ImmutableVector<T, Density>& originalVector, size_t index, const T& newValue);
    template<class T, size_t Density, class F> friend ImmutableVector<T, Density> GenerateFrom(const ImmutableVector<T, Density>& originalVector, F changeFunction);
    template<class T, size_t Density, class F> friend ImmutableVector<T, Density> GenerateNewImmutableVector(const ImmutableVector<T, Density>& originalVector, F changeFunction);
};

// Helper Functions
template<class T, size_t Density> ImmutableVector<T, Density> AppendValue(const ImmutableVector<T, Density>& originalVector, const T& newValue)
{
	const size_t newIndex = originalVector.size_;
	const size_t newSize = originalVector.size_ + 1;
	ImmutableVector<T, Density> newVector(originalVector);
	newVector.Reserve(newSize);
	++newVector.size_;
	newVector.InternalSetValue(newIndex, newValue);
	newVector.rootNode_->SetFullyConstructed();
	return newVector;
}

template<class T, size_t Density> ImmutableVector<T, Density> TruncateValue(const ImmutableVector<T, Density>& originalVector)
{
	const size_t oldIndex = originalVector.size_ - 1;
	ImmutableVector<T, Density> newVector(originalVector);
#ifdef _DEBUG
	newVector.InternalSetValue(oldIndex, T());
#endif
	--newVector.size_;
	newVector.rootNode_->SetFullyConstructed();
	return newVector;
}

template<class T, size_t Density> ImmutableVector<T, Density> SubstituteValueAtIndex(const ImmutableVector<T, Density>& originalVector, size_t index, const T& newValue)
{
	ImmutableVector<T, Density> newVector(originalVector);
	newVector.InternalSetValue(index, newValue);
	newVector.rootNode_->SetFullyConstructed();
	return newVector;
}

template<class T, size_t Density, class F> ImmutableVector<T, Density> GenerateFrom(const ImmutableVector<T, Density>& originalVector, F changeFunction)
{
	ImmutableVector<T, Density>::MutableVector newMutableVector(originalVector);
	changeFunction(newMutableVector);
	ImmutableVector<T, Density> newImmutableVector(std::move(newMutableVector));
	return newImmutableVector;
}

template<class T, size_t Density, class F> ImmutableVector<T, Density> GenerateNewImmutableVector(F changeFunction)
{
    ImmutableVector<T, Density> emptyImmutableVector;
    return GenerateFrom(emptyImmutableVector, changeFunction);
}